#include <math.h>
#include "machine.h"
#include <stdio.h>

extern double
			rosen(double *, int),
			cube(double *, int),
			beale(double *, int),
			helix(double *, int),
			powell(double *, int),
			box(double *, int),
			sing(double *, int),
			wood(double *, int),
			chebyquad(double *, int),
			watson(double *, int),
			tridiag(double *, int),
			hilbert(double *, int),
			schraube(double *, int);
extern double Praxis(double (*_fun)(double *, int), double *, double, int);
extern double tol, scbd;
extern int ktm, illc, prin, jdTest;
int ihead = 1;

struct {
	char Title[16];
	int Dimens;
	double HDist,
		Precis;
	int Start,
		ResAct;
} DTst[31] = {
	{"Rosenbrock 1", 2,  1., 1e-16, 120},
	{"Rosenbrock 2", 2,  3., 1e-16, 110},
	{"Rosenbrock 3", 2, 12., 1e-16, 206},
	{"Cube",         2,  1., 1e-16, 190},
	{"Beale",        2,  1., 1e-14, 54},
	{"Helix",        3,  1., 1e-10, 155},
	{"Powell",       3,  1., 1e-10, 55},
	{"Box",          3, 20., 1e-13, 114},
	{"Singular",     4,  1., 1e-10, 234},
	{"Wood",         4, 10., 1e-10, 496},
	{"Chebyquad 1",  2, 0.1, 1e-10, 31},
	{"Chebyquad 2",  4, 0.1, 1e-10, 74},
	{"Chebyquad 3",  6, 0.1, 1e-10, 223},
	{"Chebyquad 4",  8, 0.1, 1e-10, 326},
	{"Watson 1",     6,  1., 1e-10, 317},
	{"Watson 2",     9,  1., 1e-10, 1424},
	{"Tridiag 1",    4,  8., 1e-12, 27},
	{"Tridiag 2",    6, 12., 1e-12, 51},
	{"Tridiag 3",    8, 16., 1e-12, 123},
	{"Tridiag 4",   10, 20., 1e-12, 191},
	{"Tridiag 5",   12, 24., 1e-12, 277},
	{"Tridiag 6",   16, 32., 1e-12, 470},
	{"Tridiag 7",   20, 40., 1e-12, 788},
	{"Hilbert 1",    2, 10., 1e-14, 11},
	{"Hilbert 2",    4, 10., 1e-14, 60},
	{"Hilbert 3",    6, 10., 1e-10, 160},
	{"Hilbert 4",    8, 10., 1e-10, 409},
	{"Hilbert 5",   10, 10., 1e-10, 388},
	{"Hilbert 6",   12, 10., 1e-10, 667},
	{"Helix long",   3,  1., 1e-14, 470}
};
double xs[N];
int idiag, ireport, nTst;
FILE *filePtrD, *filePtrT;
double fnmin, diagfmin;
int n, nlSum, nfSum;
int ran2;
double ran1, ran3[127];

/*                     RANDOM NUMBER GENERATOR							*/
/*                     ***********************							*/
/*																		*/
/* Prozedure RANDOM liefert eine <Double> zufllige Zahl, die			*/
/* gleichmig in (0,1) (einschlielich 0 aber nicht 1) verteilt ist.	*/
/*     RANINIT(R) mit irgendeiner Ganzzahl R mu zur Initialisierung	*/
/* vor dem ersten Anruf der RANDOM aufgerufen werden.					*/
/* Die RAN1, RAN2 und RAN3 mssen global sein.							*/
/*     Der Algorithmus ergibt X(N)/2^56, wo								*/
/*                   X(N) = X(N-1) + X(N-127)  (mod 2^56).				*/
/* Seit 1 + X + X^127 primitiv (mod 2) ist, ist die Periode mindestens	*/
/* 2^127 - 1 > 10^38.  Siehe Knuth (1969), pp. 26, 34, 464.				*/
/*     X(N) wird in einem <Double> Wort als RAN3 = X(N)/2^56 - 1/2		*/
/* gespeichert, und gesamte Gleitkommaarithmetik ist genau.				*/

void RanInit(int ir) {
    int i, jr;
    jr = ((ir > 0)? ir: -ir) % 8190 + 1;
    ran2 = 127;
	while (ran2 > 0) {
		ran2 -= 1;  ran1 = -pow((double)2, 55);
		for (i=1; i<=7; i++) {
			jr = (1756 * jr) % 8191;
			ran1 = (ran1 + jr/32)/256;
		}
		ran3[ran2] = ran1;
	}
}
double RandBR() {
	ran2 = (ran2 == 0)? 126: ran2 - 1;
    ran1 += ran3[ran2];
	ran1 += (ran1 < 0)? 0.5: -0.5;
    ran3[ran2] = ran1;
    return ran1 + 0.5;
}

void Test(char *title, double (*_fun)(double *, int)) {
	double temp;
	n = DTst[nTst].Dimens;
	fprintf(filePtrD, "\nTest Nr  %2i  %s\n", nTst, title);
	if (nTst == 7			/* BOX		*/
     || nTst == 8			/* SINGULAR	*/
     || nTst == 9			/* WOOD		*/
     || nTst == 14			/* WATSON 1 */
	 || nTst == 15 ) {		/* WATSON 2 */
		illc = 1;
		RanInit(2);
	}
	else {
		illc = 0;
		RanInit(4);
	}
  	/* KTM for HILBERT 5 and 6 */
	ktm = (nTst == 27 || nTst == 28)? 4: 1;

    idiag = ireport = 0;	/* um das diagfmin beim 1.Mal zu fuellen */
    temp = Praxis(_fun, xs, DTst[nTst].HDist, n);
}

int main() {
   int i, ndim;
   double s1;  int s2, t1, t2;

   jdTest = 1;
   filePtrD = fopen("Diagramm.txt", "w");
   if (filePtrD == NULL) {
	   printf("Fehler beim Oeffnen der Diagramm-Datei\n");
	   return 12;
   }
   filePtrT = fopen("Report.txt", "w");
   if (filePtrT == NULL) {
	   printf("Fehler beim Oeffnen der Report-Datei\n");
	   return 12;
   }

   prin = 0;
   nlSum = nfSum = 0;
   fnmin = 0;

   nTst = 0;  xs[0] = -1.2;  xs[1] = 1.0;
   Test("Rosenbrock's function with a parabolic valley", rosen);

   nTst = 1;  xs[0] = 3;  xs[1] = 3;
   Test("Rosenbrock*s function", rosen);

   nTst = 2;  xs[0] = 8;  xs[1] = 8;
   Test("Rosenbrock's function", rosen);

   nTst = 3;  xs[0] = -1.2;  xs[1] = -1.;
   Test("Cube", cube);

   nTst = 4;  xs[0] = 0.1;  xs[1] = 0.1;
   Test("Beale", beale);

   nTst = 5;  xs[0] = -1.;  xs[1] = 0.;  xs[2] = 0.;
   Test("Helix", helix);

   nTst = 6;  xs[0] = 0.;  xs[1] = 1.;  xs[2] = 2.;
   Test("Powell", powell);

   nTst = 7;  xs[0] = 0.;  xs[1] = 10.;  xs[2] = 20.;
   Test("Box", box);

   nTst = 8;  xs[0] = 3.;  xs[1] = -1.;  xs[2] = 0.;  xs[3] = 1.;
   Test("Powell's function with a singular jacobian", sing);

   nTst = 9;   xs[0] = -3.;  xs[1] = -1.;  xs[2] = -3.;  xs[3] = -1.;
   Test("Wood", wood);

   nTst = 9;
   for (ndim=2; ndim<=8; ndim+=2) {
   	   for (i=1; i<=ndim; i++) xs[i-1] = (double)i/(ndim + 1);
	   fnmin = (ndim < 8)? 0.: 3.5168737256779e-03;
	   nTst++;  Test("Chebyquad", chebyquad);
   }

   nTst = 13;
   for (ndim=6; ndim<=9; ndim+=3) {
	   for (i=0; i<ndim; i++) xs[i] = 0.;
	   fnmin = (ndim == 6)? 0.00228767005355:
			  (ndim == 9)? 1.399760138098E-06: 0;
	   nTst++;  Test("Watson", watson);
   }

   nTst = 15;
   for (ndim=4; ndim<=20; ndim+=2) {
	   if (ndim == 14 || ndim == 18);  else {
		   for (i=0; i<ndim; i++) xs[i] = 0.;
		   fnmin = -ndim;
		   nTst++;  Test("Tridiag", tridiag);
	   }
   }

   fnmin = 0.;
   nTst = 22;
   for (ndim=2; ndim<=12; ndim+=2) {
	   for (i=0; i<ndim; i++) xs[i] = 1.;
	   nTst++;  Test("Hilbert", hilbert);
	   nTst = nTst;
   }

   nTst = 29;  xs[0] = 1.;  xs[1] = 0.;  xs[2] = 10.;
   Test("Helix long", schraube);

   s1 = 0.;  s2 = 0;
   for (i=0; i<31; i++) {
	   t1 = DTst[i].ResAct;  t2 = DTst[i].Start;
	   if (t1 > 0) {
		   s1 += (double)(t1*t1)/(t2*t2);
		   s2++;
	   }
   }
   s1 = sqrt(s1/s2);
   printf("Summary: % 11.4f\n", s1);
   fprintf(filePtrT, "Summary: % 11.4f\n", s1);

   fclose(filePtrD);
   fclose(filePtrT);
   return 0;
}

void diagramm(int nf, double fx, char *z) {
	double s;
	if (fx > fnmin) {
		if (!idiag++) diagfmin = fx;
		if (fx < diagfmin) diagfmin = fx;
		s = log(fx - fnmin);
		fprintf(filePtrD, "%04i  %+06.2f  %s\n", nf, s, z);
	}
	else {
		s = log(diagfmin - fnmin) - 2;
		fprintf(filePtrD, "%04i  %+06.2f  M\n", nf, s);
	}
}

void report(int nf, int nl, double fx, int qne, int qny) {
	double s;
	if (ihead) {
		char* head =
			"  Function        Dim   NF   NL      Fx       QNE  QNY";
		printf("%s\n", head);
		fprintf(filePtrT, "%s\n", head);
		ihead = 0;
	}
	s = fx - fnmin;
	if (!ireport && (s < DTst[nTst].Precis)) {
		printf("%-16s  %2i  %4i  %3i  % 9.2e   %2i   %2i\n",
			DTst[nTst].Title, DTst[nTst].Dimens, nf, nl, s, qne, qny);
		fprintf(filePtrT, "%-16s  %2i  %4i  %3i  % 9.2e   %2i   %2i\n",
			DTst[nTst].Title, DTst[nTst].Dimens, nf, nl, s, qne, qny);
		DTst[nTst].ResAct = nf;
		nlSum += nl;
		nfSum += nf;
		ireport = 1;
	}
}
